#include <noor_animations/animation.h>

#include <stdlib.h>
#include <stdio.h>

static void render(int timeIndex);

NoorAnimation noorAnimation = {
    "Simulation 0001",
    "Simulation animation for CS294",
    render,
    5000
};

typedef struct {
    GLfloat x, y;
} Vertex;

typedef struct {
    /* The number of vertices we have */
    size_t nVerts;

    /* Start, End, and step size */
    int startTime;
    int endTime;
    size_t stepTime;

    /* Array of vertices indexed by (t * nVerts + v) where:
     * t is the frame number in our data (ie: floor(time - startTime / frameSpacing)
     * v is the id of the vertex
     */
    Vertex *verts;

    /* Our cube */
    Vertex cube[4];
} SimData;

static SimData *simData = NULL;

void freeSimData(SimData *data) {
    if(data->verts)
        free(data->verts);
    free(data);
}

SimData *loadSimData(const char *filename) {
    FILE *fp;
    SimData *data;
    size_t i;
    int t, v;
    float x, y;

    /* Open the file */
    fp = fopen(filename, "r");
    if(!fp)
        return NULL;

    /* Allocate memory */
    data = (SimData *)calloc(1, sizeof(SimData));
    if(!data) {
        fclose(fp);
        return NULL;
    }

    /* Parse the file */
    fscanf(fp, "verts %u\n", &data->nVerts);
    fscanf(fp, "start %d\n", &data->startTime);
    fscanf(fp, "end %d\n", &data->endTime);
    fscanf(fp, "step %u\n", &data->stepTime);

    if(data->stepTime == 0) {
        fprintf(stderr, "Invalid file format: step == 0\n");
        fclose(fp);
        freeSimData(data);
        return NULL;
    }

    if(data->endTime < data->startTime) {
        fprintf(stderr, "Invalid file format: end < start\n");
        fclose(fp);
        freeSimData(data);
        return NULL;
    }

    data->verts = (Vertex *)calloc(data->nVerts * (1 + (data->endTime - data->startTime) / data->stepTime), sizeof(Vertex));
    if(!data->verts) {
        fclose(fp);
        freeSimData(data);
        return NULL;
    }

    fscanf(fp, "cube %f %f %f %f %f %f %f %f\n", &data->cube[0].x, &data->cube[0].y,
                                                 &data->cube[1].x, &data->cube[1].y,
                                                 &data->cube[2].x, &data->cube[2].y,
                                                 &data->cube[3].x, &data->cube[3].y);

/*
    fprintf(stderr, "verts %d\n", data->nVerts);
    fprintf(stderr, "start %d\n", data->startTime);
    fprintf(stderr, "end %d\n", data->endTime);
    fprintf(stderr, "step %d\n", data->stepTime);
    fprintf(stderr, "cube %f %f %f %f %f %f %f %f\n", data->cube[0].x, data->cube[0].y,
                                                      data->cube[1].x, data->cube[1].y,
                                                      data->cube[2].x, data->cube[2].y,
                                                      data->cube[3].x, data->cube[3].y);
*/

    while(fscanf(fp, "time %d vert %d %f %f\n", &t, &v, &x, &y) != EOF) {
        if(t > data->endTime) {
            fclose(fp);
            freeSimData(data);
            return NULL;
        }

        i = ((t - data->startTime) / data->stepTime) * data->nVerts + v;
        data->verts[i].x = x;
        data->verts[i].y = y;
    }

    fclose(fp);

    return data;
}

/* Render our Noor Animation at the given time index in milliseconds (10^-3)
 */
static void render(int timeIndex) {
    size_t v;
    Vertex *vPtr1;

#ifdef INTERPOLATE
    Vertex *vPtr2;
    GLfloat w1, w2;
#endif

    /* First load our simulation data */
    if(simData == NULL || timeIndex == -100) {
        if(simData != NULL) {
            freeSimData(simData);
        }
        simData = loadSimData("simulation.data");

        if(simData == NULL)
            return;
    }

    /* Set the background color to black with 0 alpha */
    glClearColor(0.0, 0.0, 0.0, 0.0);

    /* Clear the background */
    glClear(GL_COLOR_BUFFER_BIT);

    /* Make sure we're in the right matrix stack */
    glMatrixMode(GL_MODELVIEW);

    /* Draw our cube in blue*/
    glColor4f(0.1, 0.1, 0.9, 1.0);

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glBegin(GL_QUADS);

    glVertex2f(simData->cube[0].x, simData->cube[0].y);
    glVertex2f(simData->cube[1].x, simData->cube[1].y);
    glVertex2f(simData->cube[2].x, simData->cube[2].y);
    glVertex2f(simData->cube[3].x, simData->cube[3].y);

    glEnd();

    /* Draw our line in yellow */
    glColor4f(0.8, 0.9, 0.1, 1.0);

    /* Make sure our time is valid */
    if(timeIndex < simData->startTime)
        timeIndex = simData->startTime;
    if(timeIndex > simData->endTime)
        timeIndex = simData->endTime;

    vPtr1 = simData->verts + simData->nVerts * ((timeIndex - simData->startTime) / simData->stepTime);

#ifdef INTERPOLATE
    vPtr2 = vPtr1 + simData->nVerts;

    w2 = (float)((timeIndex - simData->startTime) % simData->stepTime) / (float)simData->stepTime;
    w1 = 1.0 - w2;
#endif

    glBegin(GL_LINE_STRIP);
        for(v=0; v < simData->nVerts; v++) {
#ifdef INTERPOLATE
            if(w1 == 1.0) {
                glVertex2f(vPtr1[v].x, vPtr1[v].y);
            } else {
                glVertex2f(vPtr1[v].x * w1 + vPtr2[v].x * w2,
                           vPtr1[v].y * w1 + vPtr2[v].y * w2);
            }
#else
            glVertex2f(vPtr1[v].x, vPtr1[v].y);
#endif
        }
    glEnd();
}
