/*
 * simulate 
 * argv[1] = <spring constant>
 * argv[2] = <rest length> 
 * argv[3] = <mass>
 * argv[4-6] = <anchor x, y, z>
 * argv[7-9] = <initial pos>
 * argv[10-12] = <initial velocity>
 * argv[13-15] = <grav accel>
 * argv[16] = <efficiency>
 * argv[17] = <start frame> (corresponds to initial pos/vel)
 * argv[18] = <end frame>
 * argv[19] = <steps per frame>
 * argv[20] = <fps>
 * argv[21] = <save file>
 *
 */

#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include <ode/solver.h>
#include <ode/step_euler.h>
#include <ode/step_midpoint.h>

/* **** VECTORS **** */
typedef struct {
    ODEFloat x, y, z;
} vector_t;

typedef struct {
    ODEFloat k;
    ODEFloat lRest;
    ODEFloat mass;
    vector_t anchor;
    vector_t aGrav;
    ODEFloat efficiency;
    FILE *fp;

    ODEFloat last_t;

    int initialized; /* Boolean flag for copy optimization.
                        We only copy static data if this is false. */
} SimData;

/* Magnitude and dot product are simple enough to be macros */
#define dot(a,b)    ((a).x * (b).x + (a).y * (b).y + (a).z * (b).z)
#define vector_mag(v)    sqrt(dot(v,v))

inline vector_t vector_add(vector_t v1, vector_t v2) {
    v1.x += v2.x;
    v1.y += v2.y;
    v1.z += v2.z;

    return v1;
}

inline vector_t vector_sub(vector_t v1, vector_t v2) {
    v1.x -= v2.x;
    v1.y -= v2.y;
    v1.z -= v2.z;

    return v1;
}

inline vector_t vector_scale(ODEFloat s, vector_t v) {
    v.x *= s;
    v.y *= s;
    v.z *= s;

    return v;
}

inline vector_t unit_vector(vector_t v) {
    return vector_scale(1.0 / vector_mag(v), v);
}

/* What is thr force at the end point */
vector_t spring_force(vector_t anchor, vector_t end, ODEFloat k, ODEFloat lRest) {
    vector_t deltaP = vector_sub(end, anchor);

    /* how much are we stretched? */
    ODEFloat l = vector_mag(deltaP);

    /* So what is the force */
    ODEFloat force = (lRest - l) * k;

    /* And return it in vector form */
    return vector_scale(force, unit_vector(deltaP));
}

static void f(const ODEData *data, ODEFloat *f_result) {
    SimData *simData = (SimData *)data->simData;
    vector_t pos = *((vector_t *)data->x);
    vector_t fSpring = spring_force(simData->anchor, pos, simData->k, simData->lRest);
    vector_t aSpring = vector_scale(1.0 / simData->mass, fSpring);

    /* dx/dt = vel */
    f_result[0] = data->x[3];
    f_result[1] = data->x[4];
    f_result[2] = data->x[5];    

    /* dv/dt = accel */
    f_result[3] = aSpring.x + simData->aGrav.x;
    f_result[4] = aSpring.y + simData->aGrav.y;
    f_result[5] = aSpring.z + simData->aGrav.z;

    fprintf(stderr, "t=%.2f p=<%.2f, %.2f, %.2f> v=<%.2f, %.2f, %.2f> a=<%.2f, %.2f, %.2f>\n", data->t, data->x[0], data->x[1], data->x[2], data->x[3], data->x[4], data->x[5], f_result[3], f_result[4], f_result[5]);
}

static void reportCB(const ODEData *data) {
    SimData *simData = (SimData *)data->simData;
    fprintf(simData->fp, "%f %f %f\n", data->x[0], data->x[1], data->x[2]);
}

static void simDataUpdateCB(ODEData *data) {
    SimData *simData = (SimData *)data->simData;

    /* efficiency is the ammount of energy conserved over one second
     * disceff is the ammount of energy conserved over one frame
     *
     * stepeff^(1.0 / h) = efficiency
     * => stepeff = efficiency^h
     */
    ODEFloat stepeff = pow(simData->efficiency, data->t - simData->last_t);

    vector_t vel = *((vector_t *)(data->x + 3));

    /* Dampening: The percentage of energy conserved is equal to stepeff */
    ODEFloat energy = vector_mag(vel);
    energy *= energy;  /* Current energy : K proportional to v^2 */
    energy *= stepeff; /* Factor in energy loss */

    if(energy > 0.0) {
        vel = vector_scale(sqrt(energy), unit_vector(vel)); /* Rescale the vector */
    } else {
        vel.x = vel.y = vel.z = 0.0;
    }

    data->x[3] = vel.x;
    data->x[4] = vel.y;
    data->x[5] = vel.z;

    simData->last_t = data->t;
}

/* Allocate our simulation data */
void *simDataAllocateCB() {
    SimData *data = (SimData *)malloc(sizeof(SimData));

    /* Make sure we got the space */
    if(data)
        data->initialized = 0;

    return data;
}

/* Copy our simulation data */
void simDataCopyCB(void *dest, const void *src) {
    SimData *sdest = (SimData *)dest;
    SimData *ssrc = (SimData *)src;

    if(! sdest->initialized) {
        *sdest = *ssrc;
    }
}

/*
 * simulate 
 * argv[1] = <spring constant>
 * argv[2] = <rest length> 
 * argv[3] = <mass>
 * argv[4-6] = <anchor x, y, z>
 * argv[7-9] = <initial pos>
 * argv[10-12] = <initial velocity>
 * argv[13-15] = <grav accel>
 * argv[16] = <efficiency>
 * argv[17] = <start frame> (corresponds to initial pos/vel)
 * argv[18] = <end frame>
 * argv[19] = <steps per frame>
 * argv[20] = <fps>
 * argv[21] = <save file>
 *
 */

int main(int argc, char **argv) {
    SimData *simData = simDataAllocateCB();

    /* Assume we're called intelligently */
    /* Passed Arguments */
    simData->k = atof(argv[1]);
    simData->lRest = atof(argv[2]);
    simData->mass = atof(argv[3]);
    simData->anchor.x = atof(argv[4]);
    simData->anchor.y = atof(argv[5]);
    simData->anchor.z = atof(argv[6]);
    vector_t pInitial = {atof(argv[7]), atof(argv[8]), atof(argv[9])};
    vector_t vInitial = {atof(argv[10]), atof(argv[11]), atof(argv[12])};
    simData->aGrav.x = atof(argv[13]);
    simData->aGrav.y = atof(argv[14]);
    simData->aGrav.z = atof(argv[15]);
    simData->efficiency = atof(argv[16]);
    int32_t startFrame= atoi(argv[17]);
    int32_t endFrame = atoi(argv[18]);
    ODEFloat steps = atof(argv[19]);
    ODEFloat framerate = atof(argv[20]);
    const char *cacheFile = argv[21];

    /* Compute these given the old-style inputs */
    ODEFloat endTime = (ODEFloat)(endFrame - startFrame) / framerate;
    ODEFloat reportFreq = 1.0 / framerate;
    ODEFloat stepSize = reportFreq / steps;

    /* Leave negative steps untouched */
    if(steps < 0.0)
        stepSize = steps;

    /* Set our initial configuration */
    simData->last_t = 0.0;

    ODEData *data = (ODEData *)alloca(sizeof(ODEData));
    data->x = (ODEFloat *)alloca(sizeof(ODEFloat) * 6);
    data->x[0] = pInitial.x;
    data->x[1] = pInitial.y;
    data->x[2] = pInitial.z;
    data->x[3] = vInitial.x;
    data->x[4] = vInitial.y;
    data->x[5] = vInitial.z;
    data->dim = 6;
    data->t = 0.0;

    /* Open our output file */
    simData->fp = fopen(cacheFile, "w");
    if(!simData->fp) {
        fprintf(stderr, "Can't open %s.  Using stdout.\n", cacheFile);
        simData->fp = stdout;
    }

    /* Store our simulation data */
    simData->initialized = 1;
    data->simData = (void *)simData;

    /* Run the solver */
    ode_solver(f, data, ode_step_midpoint, reportCB, simDataUpdateCB, simDataAllocateCB, free, simDataCopyCB, reportFreq, endTime, stepSize);

    /* Close the output file */
    if(simData->fp != stdout)
        fclose(simData->fp);

    return 0;
}
