#include <string.h>
#include <stdio.h>

#include "solver.h"

#define step(s, e, h) stepFunc(f, s, e, h, simDataUpdateCB, simDataAllocateCB, simDataFreeCB, simDataCopyCB, &stepData, &stepDataFreeCB)

int ode_solver(ODEFunc f, ODEData *data, ODEStepFunc stepFunc,
               ODEReportCB reportCB, ODESimDataUpdateCB simDataUpdateCB,
               ODESimDataAllocateCB simDataAllocateCB,
               ODESimDataFreeCB simDataFreeCB, ODESimDataCopyCB simDataCopyCB,
               ODEFloat reportFreq, ODEFloat endTime, ODEFloat stepSize) {
    ODEFloat t_reported; /* Our last reported time */
    ODEFloat h;          /* Our step size in t... ie delta-t */
    int adaptive;        /* Boolean... is our stepSize adaptive? */
    int refine;          /* Boolean... do we need to refine our stepsize */
    ODEFloat threshold;  /* The adaptive threshold */
    ODEFloat e;          /* Current error */

    ODEData *data_mid  = NULL;   /* Temporary values for adaptive time stepping */
    ODEData *data_end  = NULL;   /* Temporary values for adaptive time stepping */
    ODEData *data_end2 = NULL;   /* Temporary values for adaptive time stepping */

    size_t i;

    void *stepData = NULL;
    ODEStepDataFreeCB stepDataFreeCB = NULL;

    /* Simple error checking */
    if(!f || !data || !data->x || !stepFunc || !reportCB || reportFreq <= 0.0
       || stepSize == 0.0
       || (simDataUpdateCB && (!simDataAllocateCB || !simDataFreeCB || !simDataCopyCB))
       || (simDataAllocateCB && (!simDataUpdateCB || !simDataFreeCB || !simDataCopyCB))
       || (simDataFreeCB && (!simDataAllocateCB || !simDataUpdateCB || !simDataCopyCB))
       || (simDataCopyCB && (!simDataAllocateCB || !simDataUpdateCB || !simDataUpdateCB)))
        return 1;

    /* Always report our starting point */
    reportCB(data);
    t_reported = data->t;

    /* Figure out our first step size */
    adaptive = (stepSize < 0.0);

    /* If we're adaptive, set the first stepsize guess to be the reporting frequency. */
    if(!adaptive) {
        h = stepSize;
    } else {
        threshold = -stepSize;
        h = reportFreq;

        data_mid  = (ODEData *)alloca(sizeof(ODEData));
        data_end  = (ODEData *)alloca(sizeof(ODEData));
        data_end2 = (ODEData *)alloca(sizeof(ODEData));

        if(!data_mid || !data_end || !data_end2)
            return 2;

        data_mid->dim  = data->dim;
        data_end->dim  = data->dim;
        data_end2->dim = data->dim;

        data_mid->x  = (ODEFloat *)alloca(data->dim * sizeof(ODEFloat));
        data_end->x  = (ODEFloat *)alloca(data->dim * sizeof(ODEFloat));
        data_end2->x = (ODEFloat *)alloca(data->dim * sizeof(ODEFloat));

        if(!data_mid->x || !data_end->x || !data_end2->x)
            return 2;

        /* We need simData at the midpoint, since we step from there.  */
        data_mid->simData  = simDataAllocateCB();

        if(!data_mid->simData)
            return 2;
    }

    while(t_reported < endTime) {
        if(adaptive) {
            /* Set to our report time if neccessary */
            if (data->t - t_reported + h > reportFreq) {
                h = t_reported + reportFreq - data->t;
            }

            refine =  1;
            while(refine) {
                step(data, data_end, h);
                step(data, data_mid, h / 2.0);

                /* Update the midpoint's simulation data, so we can take a step from it */
                if(simDataCopyCB) {
                    simDataCopyCB(data_mid->simData, data->simData);
                    simDataUpdateCB(data_mid);
                }

                step(data_mid, data_end2, h / 2.0);

                /* See if we need to refine any more */
                refine = 0;
                for(i=0; i < data->dim; i++) {
                    e = (data_end->x[i] - data_end2->x[i]);
                    if((e > 0.0 ? e : -e) > threshold) {
                        refine = 1;
                        h /= 2.0;
                        break;
                    }
                }
            }

            /* Copy data from data_end2 to data */
            data->t = data_end2->t;
            /*  for(i=0; i < data->dim; i++) {
             *      data->x[i] = data_end2->x[i];
             *  }
             */
            memcpy(data->x, data_end2->x, sizeof(ODEFloat) * data->dim);

            /* Grow the timestep */
            h *= 2.0;
        } else {
            step(data, data, h);
        }

        /* Let the simulation update its internal data for the new state */
        if(simDataUpdateCB)
            simDataUpdateCB(data);

        /* Report if neccessary - be careful of flop precision */
        if(data->t - t_reported - reportFreq > reportFreq * -0.001) {
            reportCB(data);
            t_reported = data->t;
        }
    }

    /* Free heap space we've allocated */
    if(adaptive)
        simDataFreeCB(data_mid->simData);

    if(stepData)
        stepDataFreeCB(stepData);

    return 0;
}
