#ifndef _ODE_SOLVER_H_
#define _ODE_SOLVER_H_

#include "types.h"

/* ode_solver solves an arbitrary ODE.  It takes the following parameters:
 *   - The ODEFunction: f(x,t)
 *   - The initial ODEData corresponding to t=0
 *   - The step function
 *   - The reporting callback
 *   - The functions to update, allocate, free, and copy simData
 *   - The reporting frequency (we use the report callback at this interval.
 *     note that as t gets large, we loose fidelity doe to flop precision)
 *   - The stopping time
 *   - The step size (a negative amount means adaptive stepping with the
 *     magnitude being the acceptable error)
 *
 *   NOTE: If step size is not adaptive, the reportCB will be called once the
 *         reporting time is hit or passed.  Thus because of floating point
 *         precision problems, the report might not happen at the exact desired
 *         interval even if the report frequency is an integer multiple of the
 *         step size.  If the step size is adaptive, the times of the reports
 *         should be as on target as floating point precision allows.
 *
 *   NOTE: The initial config and config corresponding to the end time are
 *         always reported.
 *
 *   NOTE: Currently, we are assuming the dimensionality doesn't change during
 *         the sim.
 *
 *   NOTE: The simData pointer in the initial data might not always be the same pointer
 *         throughout the simulation, but the pointer will always reference valid data.
 *         when simDataUpdateCB is called, the simData will always be for the most recent
 *         timestep before the current time.
 *
 *   Return values:
 *   0 - success
 *   1 - invalid input
 *   2 - memory allocation error
 */

int ode_solver(ODEFunc f, ODEData *data, ODEStepFunc stepFunc,
               ODEReportCB reportCB, ODESimDataUpdateCB simDataUpdateCB,
               ODESimDataAllocateCB simDataAllocateCB,
               ODESimDataFreeCB simDataFreeCB, ODESimDataCopyCB simDataCopyCB,
               ODEFloat reportFreq, ODEFloat endTime, ODEFloat stepSize);

#endif /* _ODE_SOLVER_H_ */
