#include "step_midpoint.h"

#include <stdlib.h>

typedef struct {
    ODEFloat *f_result;

    ODEData *end1;

    void *simData;
    ODESimDataFreeCB simDataFreeCB;
} ODEStepMidpointData;

static void freeMidpointData(void *data) {
    ODEStepMidpointData *sdata = (ODEStepMidpointData *)data;

    if(sdata->f_result)
        free(sdata->f_result);

    if(sdata->simData)
        sdata->simDataFreeCB(sdata->simData);
}

/* Explicit Integrator Step Function: Midpoint Method */
int ode_step_midpoint(ODEFunc f, const ODEData *start, ODEData *end,
                      ODEFloat h, ODESimDataUpdateCB simDataUpdateCB,
                      ODESimDataAllocateCB simDataAllocateCB,
                      ODESimDataFreeCB simDataFreeCB, ODESimDataCopyCB simDataCopyCB,
                      void **_stepData, ODEStepDataFreeCB *stepDataFreeCB) {
    size_t i;
    ODEStepMidpointData *stepData = *_stepData;
    ODEFloat *f_result_start;
    ODEFloat *f_result_end;
    ODEFloat *x_end1;
    ODEFloat *saveEndX;
    void *saveEndSimData = NULL;

    /* Allocate our space if needed */
    if(! stepData) {
        stepData = malloc(sizeof(ODEStepMidpointData));

        if(!stepData)
            return 2;

        stepData->f_result = (ODEFloat *)malloc(3 * start->dim * sizeof(ODEFloat));

        if(!stepData->f_result) {
            free(stepData);
            return 2;
        }

        stepData->simDataFreeCB = simDataFreeCB;

        *_stepData = stepData;
        *stepDataFreeCB = freeMidpointData;
    }

    f_result_start = stepData->f_result;
    f_result_end = f_result_start + start->dim;
    x_end1 = f_result_end + start->dim;

    /* Evaluate the differential equation with the current state to get dx/dt */
    f(start, f_result_start);

    /* Scale the results by delta-t so we get delta-x */
    for(i=0; i < start->dim; i++) {
        x_end1[i] = start->x[i] + h * f_result_start[i];
    }

    /* Update t */
    end->t = start->t + h;

    /* If we are using simulation data, we need to update it */
    if(start->simData) {
        /* Store the end pointer to restore later */
        saveEndSimData = end->simData;

        /* First we need to make sure we have our space to make the backup */
        if(!stepData->simData) {
            stepData->simData = simDataAllocateCB();

            if(!stepData->simData)
                return 2;
        }

        simDataCopyCB(stepData->simData, start->simData);

        end->simData = stepData->simData;
        simDataUpdateCB(end);
    }

    /* Backup end->x and set it to our temp values */
    saveEndX = end->x;
    end->x = x_end1;

    /* Evaluate the differential equation at the end of the euler step */
    f(end, f_result_end);

    /* Restore end->x */
    end->x = saveEndX;

    /* Restore simData */
    if(start->simData)
        end->simData = saveEndSimData;

    /* Choose the median dx/dt and scale the results by delta-t so we get delta-x */
    for(i=0; i < start->dim; i++) {
        end->x[i] = start->x[i] + h * (f_result_start[i] + f_result_end[i]) / 2.0;
    }

    return 0;
}
