#ifndef _ODE_TYPES_H_
#define _ODE_TYPES_H_

#include <stdlib.h>

/* Set our floating point precision */
/* typedef float ODEFloat; */
typedef double ODEFloat;

/* The current state */
typedef struct {
    ODEFloat *x;
    ODEFloat t;
    size_t dim;

    /* This following pointer is not used directly by the solver.  It is
     * provided to allow the simulator to maintain its own data to be used
     * by ODEFunc.  After x and t are updated (ie after each timestep or
     * during temporary stages by the step function), the ODESimDataUpdateCB
     * callback will be called to allow the simulator to update this data.
     */
    void *simData;
} ODEData;

/* ODEFunc is f(x, t-current)... dx/dt evaluated at x, t-current.
 * The caller (solver) is responsible for allocating the array into which  the
 * result will be put.  It should have data->dim slots.
 */
typedef void (*ODEFunc)(const ODEData *, ODEFloat *);

/* ODEReportCB is a typedef for a callback that is used to report the state of
 * the system at regular intervals
 */
typedef void (*ODEReportCB)(const ODEData *);

/* See the comment above simData in the ODEData typedef for information on
 * ODESimDataUpdateCB.
 */
typedef void (*ODESimDataUpdateCB)(ODEData *);

/* The solver or step function may need to allocate, free, or copy simData.
 * The following callbacks are used by the system to do these tasks.
 */
typedef void *(*ODESimDataAllocateCB)();
typedef void (*ODESimDataFreeCB)(void *);
typedef void (*ODESimDataCopyCB)(void *, const void *);

/* ODEStepFunc is a typedef for a function to step an ODE.  A function of this
 * type is passed to the solver and is used to solve the ODE.  Many step
 * functions are provided by the library and are available in step_*.h.
 * Additionally, it's possible to use your write your own step function.
 * A step function takes the following arguments:
 *   - The ODEFunction: f(x,t)
 *   - The start state
 *   - The end state
 *   - The step size (h, aka delta-t)
 *   - Functions for allocating, freeing, and copying simulation data
 *   - pointer to a pointer to persistent data for this step function
 *   - pointer to a function to free this persistent data
 *
 * NOTE on persistent data:
 *   The solver maintains a pointer to persistent data for the step function.
 *   This pointer is initially NULL.  It is passed as a reference to the step
 *   function, and it is the step function's responsibility to allocate space
 *   if it needs it and report back to the solver how to free this space when
 *   the solver is done.  To do this, the step function sets the free function
 *   pointer to point to the function to free this data (eg:
 *   *stepDataFreeCB = free;).  This should be used to maintain space for
 *   storing temporary results (like f(x,t), midpoint values, etc).
 *
 * NOTE: Functions of this type must be able to handle the case where
 *       end = start (ie, we're overwriting the initial data)
 *
 * Assumtions this function can make:
 *   - end->dim is set properly
 *   - end->x is allocated
 *   - f_result is allocated and has start->dim buckets
 *   - ODESimDataUpdateCB ODESimDataAllocateCB ODESimDataFreeCB
 *     ODESimDataCopyCB are all NULL or all non-null.  If null, then there is
 *     no simData.  These methods are provided to the step function so it can
 *     generate valid states if neccessary, but when the step function returns,
 *     start->simData and end->simData should reference the same location, and
 *     the data should be unchanged.
 *
 *   Return values:
 *   0 - success
 *   1 - invalid input
 *   2 - memory allocation error
 */
typedef void (*ODEStepDataFreeCB)(void *);
typedef int (*ODEStepFunc)(ODEFunc, const ODEData *, ODEData *, ODEFloat,
                           ODESimDataUpdateCB, ODESimDataAllocateCB,
                           ODESimDataFreeCB, ODESimDataCopyCB, void **,
                           ODEStepDataFreeCB *);

#endif /* _ODE_TYPES_H_ */
