#include <noor_animations/animation.h>
#include <noor_animations/nautil.h>
#include <pgl/pgl.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static void render(int timeIndex);

NoorAnimation noorAnimation = {
    "Plinko 0001",
    "Plinko Simulation Animation For CS294-1",
    render,
    10000
};

typedef struct {
    GLfloat x, y;
} Vertex;

typedef struct {
    Vertex *verts;
    size_t nVerts;
} Polygon;

typedef struct {
    Vertex center;
    GLfloat r;
} Circle;

typedef struct {
    Vertex v1;
    Vertex v2;
} Line;

#define SHAPE_POLYGON 1
#define SHAPE_CIRCLE  2
#define SHAPE_LINE    3

typedef struct {
    uint32_t type;

    union {
        Polygon p;
        Circle c;
        Line l;
    } geom;

    NAColor color;
} Shape;

/* All balls are the same size/color */
typedef struct {
    int visible;
    Vertex center;
} Ball;

typedef struct {
    /* The number of static geometry we have */
    size_t nStatic;

    /* The number of balls we have */
    size_t nBalls;

    /* Our timing information */
    int startTime;
    int endTime;
    int stepTime;

    /* Array of static shapes in our scene */
    Shape *staticShapes;

     /* Array of balls indexed by (t * nVerts + b) where:
      * t is the frame number in our data (ie: floor(time - startTime / frameSpacing)
      * b is the id of the ball
      */
    Ball *balls;
} SimData;

static SimData *simData = NULL;

void freeSimData(SimData *data) {
    size_t i;

    if(data->staticShapes) {
        for(i=0; i < data->nStatic; i++) {
            if(data->staticShapes[i].type == SHAPE_POLYGON && 
               data->staticShapes[i].geom.p.verts)
            free(data->staticShapes[i].geom.p.verts);
        }
        free(data->staticShapes);
    }

    if(data->balls)
        free(data->balls);
    free(data);
}

SimData *loadSimData(const char *filename) {
    FILE *fp;
    SimData *data;
    size_t i, j;
    int t, b, v;
    float x, y;
    char buffer[2048];

    /* Open the file */
    fp = fopen(filename, "r");
    if(!fp)
        return NULL;

    /* Allocate memory */
    data = (SimData *)calloc(1, sizeof(SimData));
    if(!data) {
        fclose(fp);
        return NULL;
    }

    /* Parse the file */
    fscanf(fp, "static %u\n", (unsigned int *)&data->nStatic);
    fscanf(fp, "balls %u\n", (unsigned int *)&data->nBalls);
    fscanf(fp, "start %d\n", &data->startTime);
    fscanf(fp, "end %d\n", &data->endTime);
    fscanf(fp, "step %u\n", &data->stepTime);

    noorAnimation.length = data->endTime - data->startTime;

    if(data->stepTime == 0) {
        fprintf(stderr, "Invalid file format: step == 0\n");
        fclose(fp);
        freeSimData(data);
        return NULL;
    }

    if(data->endTime < data->startTime) {
        fprintf(stderr, "Invalid file format: end < start\n");
        fclose(fp);
        freeSimData(data);
        return NULL;
    }

    data->balls = (Ball *)calloc(data->nBalls * (1 + (data->endTime - data->startTime) / data->stepTime), sizeof(Ball));
    if(!data->balls) {
        fclose(fp);
        freeSimData(data);
        return NULL;
    }

    data->staticShapes = (Shape *)calloc(data->nStatic, sizeof(Shape));
    if(!data->staticShapes) {
        fclose(fp);
        freeSimData(data);
        return NULL;
    }

    for(i=0; i < data->nStatic; i++) {
        fscanf(fp, "%s color %f %f %f %f\n", buffer, &data->staticShapes[i].color.r, &data->staticShapes[i].color.g, &data->staticShapes[i].color.b, &data->staticShapes[i].color.a);

        if(strcmp(buffer, "polygon") == 0) {
            data->staticShapes[i].type = SHAPE_POLYGON;
            fscanf(fp, "\tverts %u\n", (unsigned int *)&data->staticShapes[i].geom.p.nVerts);

            /* Allocate space for the vertices */
            data->staticShapes[i].geom.p.verts = (Vertex *)malloc(data->staticShapes[i].geom.p.nVerts * sizeof(Vertex));
            if(!data->staticShapes[i].geom.p.verts) {
                fclose(fp);
                freeSimData(data);
                return NULL;
            }

            for(j=0; j <= data->staticShapes[i].geom.p.nVerts; j++) {
                fscanf(fp, "\tvert %f %f\n", &data->staticShapes[i].geom.p.verts[j].x, &data->staticShapes[i].geom.p.verts[j].y);
            }
        } else if(strcmp(buffer, "circle") == 0) {
            data->staticShapes[i].type = SHAPE_CIRCLE;
            fscanf(fp, "\tcenter %f %f\n", &data->staticShapes[i].geom.c.center.x, &data->staticShapes[i].geom.c.center.y);
            fscanf(fp, "\tradius %f\n", &data->staticShapes[i].geom.c.r);
        } else if(strcmp(buffer, "line") == 0) {
            data->staticShapes[i].type = SHAPE_LINE;
            fscanf(fp, "\tvert %f %f\n", &data->staticShapes[i].geom.l.v1.x, &data->staticShapes[i].geom.l.v1.y);
            fscanf(fp, "\tvert %f %f\n", &data->staticShapes[i].geom.l.v2.x, &data->staticShapes[i].geom.l.v2.y);
        }
    }

    while(fscanf(fp, "time %d ball %d %d %f %f\n", &t, &b, &v, &x, &y) != EOF) {
        if((t > data->endTime + data->stepTime / 2) || b >= data->nBalls)
            continue;

        i = ((t - data->startTime) / data->stepTime) * data->nBalls + b;

        /* Handle flop related issues in the sim where the time might be slightly
         * below what we expect... ie. round t to the nearest valid value.
         */
        if(((t - data->startTime) % data->stepTime) > (data->stepTime >> 1))
            i += data->nBalls;

        data->balls[i].visible = v;
        data->balls[i].center.x = x;
        data->balls[i].center.y = y;
    }

    fclose(fp);
    return data;
}

/* Render our Noor Animation at the given time index in milliseconds (10^-3)
 */
static void render(int timeIndex) {
    size_t i, j, b;
    Ball *bPtr1;

#ifdef INTERPOLATE
    Ball *bPtr2;
    GLfloat w1, w2;
#endif

    /* First load our simulation data */
    if(simData == NULL || timeIndex == -100) {
        if(simData != NULL) {
            freeSimData(simData);
        }
        simData = loadSimData("simulation.data");

        if(simData == NULL)
            return;
    }

    /* Set the background color to black with 0 alpha */
    glClearColor(0.0, 0.0, 0.0, 0.0);

    /* Clear the background */
    glClear(GL_COLOR_BUFFER_BIT);

    /* Make sure we're in the right matrix stack */
    glMatrixMode(GL_MODELVIEW);

    /* Draw our static elements */
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    for(i=0; i < simData->nStatic; i++) {
        naColor(simData->staticShapes[i].color);

        switch(simData->staticShapes[i].type) {
            case SHAPE_POLYGON:
                glBegin(GL_POLYGON);
                for(j=0; j < simData->staticShapes[i].geom.p.nVerts; j++) {
                    glVertex2f(simData->staticShapes[i].geom.p.verts[j].x, simData->staticShapes[i].geom.p.verts[j].y);
                }
                glEnd();
                break;
            case SHAPE_CIRCLE:
                glPushMatrix();
                glTranslatef(simData->staticShapes[i].geom.c.center.x, simData->staticShapes[i].geom.c.center.y, 0.0);
                pglCirclef(0.0, 0.0, simData->staticShapes[i].geom.c.r);
                glPopMatrix();
                break;
            case SHAPE_LINE:
                glBegin(GL_LINES);
                glVertex2f(simData->staticShapes[i].geom.l.v1.x, simData->staticShapes[i].geom.l.v1.y);
                glVertex2f(simData->staticShapes[i].geom.l.v2.x, simData->staticShapes[i].geom.l.v2.y);
                glEnd();
                break;
            default:
                break;
        }
    }

    /* Draw our balls in white */
    glColor4f(0.8, 0.8, 0.8, 1.0);

    /* Make sure our time is valid */
    if(timeIndex < simData->startTime)
        timeIndex = simData->startTime;
    if(timeIndex > simData->endTime)
        timeIndex = simData->endTime;

    bPtr1 = simData->balls + simData->nBalls * ((timeIndex - simData->startTime) / simData->stepTime);

#ifdef INTERPOLATE
    bPtr2 = bPtr1 + simData->nBalls;

    w2 = (float)((timeIndex - simData->startTime) % simData->stepTime) / (float)simData->stepTime;
    w1 = 1.0 - w2;
#endif

    for(b=0; b < simData->nBalls; b++) {
        if(!bPtr1[b].visible)
            continue;

#define drawBall(x, y) glPushMatrix(); glTranslatef(x, y, 0.0); pglCirclef(0.0, 0.0, 0.07); glPopMatrix();

#ifdef INTERPOLATE
        if(w1 == 1.0 || !bPtr2[b].visible) {
            drawBall(bPtr1[b].center.x, bPtr1[b].center.y);
        } else {
            drawBall(bPtr1[b].center.x * w1 + bPtr2[b].center.x * w2,
                     bPtr1[b].center.y * w1 + bPtr2[b].center.y * w2);
        }
#else
        drawBall(bPtr1[b].center.x, bPtr1[b].center.y);
#endif
    }
}
