/* This is a brute-force rigid body simulation of plinko */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <ode/solver.h>
#include <ode/step_euler.h>
#include <ode/step_midpoint.h>

/* ground is the absolute ground.  bottom is where the buckets start */
#define ground -0.95
#define top 1.1
#define bottom -0.7
#define left -0.95
#define right 0.95

#define pegRadius 0.025
#define ballRadius 0.07
#define contactEpsilon 0.0005
#define contactEpsilonSq (contactEpsilon*contactEpsilon)

/* Keep pegRows odd to make the math easy */
#define pegRows 9

/* Keep this even. The actual # alternates between pegCols and pegCols-1.
 * The first and last row have pegCols-1 pegs
 */
#define pegCols 8

#define nPegs (pegRows * pegCols - (pegRows >> 1) - 1)

#define numBalls 6

#define pegElasticity 0.2
#define ballElasticity 0.4

typedef struct _Vector {
    ODEFloat x, y;
} Vector;

/* Magnitude and dot product are simple enough to be macros */
#define dot(a,b)    ((a).x * (b).x + (a).y * (b).y)
#define vector_mag(v) sqrt(dot(v,v))
#define dSquared(v1, v2) (((v1).x - (v2).x) * ((v1).x - (v2).x) + ((v1).y - (v2).y) * ((v1).y - (v2).y))

static inline Vector vector_add(Vector v1, Vector v2) {
    v1.x += v2.x;
    v1.y += v2.y;

    return v1;
}

static inline Vector vector_sub(Vector v1, Vector v2) {
    v1.x -= v2.x;
    v1.y -= v2.y;

    return v1;
}

static inline Vector vector_scale(ODEFloat s, Vector v) {
    v.x *= s;
    v.y *= s;

    return v;
}

static inline Vector unit_vector(Vector v) {
    return vector_scale(1.0 / vector_mag(v), v);
}

/* So we can typecase (x + b * 5) to something useful */
typedef struct _Ball {
    ODEFloat visible;
    Vector center;
    Vector velocity;
} Ball;

typedef struct _SimData {
    /* The time of our last update */
    ODEFloat tUpdated;

    /* The initial state of the balls with visible being the spawn time
     * and center/velocity being initial conditions at that time.
     */
    Ball *initialBalls;
    size_t nBalls;

    Ball *lastBalls;

    /* The next ball to come alive */
    size_t nextActive;
} SimData;

/* The last peg row starts the collectors */
static Vector pegs[pegRows][pegCols];

/* Report the state of the system (ie print our frame) */
static void report(const ODEData *data) {
    SimData *simData = (SimData *)data->simData;
    Ball *balls = (Ball *)data->x;
    size_t i;

    for(i=0; i < simData->nBalls; i++) {
        printf("time %d ball %d %d %f %f\n", (int)(data->t * 1000.0), (int)i,
               (int)balls[i].visible, balls[i].center.x, balls[i].center.y);
    }
}

static inline void reflect(Vector I, Vector nCol, Vector pCol, ODEFloat tCol, SimData *simData, ODEData *data, size_t b) {
    ODEFloat tPostCol;
    Ball *balls = (Ball *)data->x;
    Vector ref = vector_sub(I, vector_scale(2 * dot(nCol, I), nCol));
    
    /* Compute the new velocity */
    ODEFloat energy = pegElasticity * (1.0 - dot(I, nCol)) *
                      dot(balls[b].velocity, balls[b].velocity);
    balls[b].velocity = vector_scale(sqrt(energy), ref);
    
    /* Use that velocity to compute the new location */
    tPostCol = (data->t - simData->tUpdated) * tCol;
    balls[b].center = vector_scale(tPostCol, balls[b].velocity);
    balls[b].center = vector_add(pCol, balls[b].center);
    
    /* Update the start location, so we have a valid path if we have multiple
     * collisions in one time step.
     */
    simData->lastBalls[b].velocity = balls[b].velocity;
    tPostCol = (data->t - simData->tUpdated) * (1.0 - tCol);
    simData->lastBalls[b].center = vector_scale(-tPostCol, balls[b].velocity);
}

/* Update the simulation data.  This is where we handle collisions, spawn
 * new balls, etc. */
static void simDataUpdate(ODEData *data) {
    SimData *simData = (SimData *)data->simData;
    Ball *balls = (Ball *)data->x;
    size_t i, r, c;
    int w;
    ODEFloat colDistSq, dSq;
    ODEFloat tCol;
    ODEFloat tLow, tHigh, tLowDSq, tHighDSq;
    Vector pStart, pCol, nCol, v, pTemp;

    /* fprintf(stderr, "simDataUpdate: %f\n", data->t); */

    /* Check ball/peg collision */
    colDistSq = pegRadius + ballRadius;
    colDistSq *= colDistSq;
    for(i=0; i < simData->nextActive; i++) {
        for(r=0; r < pegRows; r++) {
            for(c=0; c < pegCols; c++) {
                /* Even rows don't have the last peg */
                if((r % 2 == 0) && c == (pegCols-1))
                    continue;

                dSq = dSquared(balls[i].center, pegs[r][c]);
                if(dSq < colDistSq) {
                    /* We have a collision */

                    /* TODO: Add a constraint that will affect f(x, t) */

                    /* Find our path over this time step */
                    pStart = simData->lastBalls[i].center;
                    v = vector_sub(balls[i].center, pStart);

                    /* Intelligent-Binary search to find the point of collision */
                    tHigh = 1.0;
                    tHighDSq = dSq;
                    tLow = 0.0;
                    tLowDSq = dSquared(pStart, pegs[r][c]);

                    /* DEBUGGING: started step interconnected */
                    if(tLowDSq < colDistSq) {
                        fprintf(stderr, "Collision (peg) detected at start of timestep.\n");
                    }

                    /*
                     * fprintf(stderr, "collision t=[%.3f,%.3f]: ball %d (%.3f, %.3f), peg %d,%d (%.3f, %.3f)\n", simData->tUpdated, data->t, (int)i, balls[i].center.x, balls[i].center.y, (int)r, (int)c, pegs[r][c].x, pegs[r][c].y);
                     * fprintf(stderr, "    dSq=%f, colDistSq=%f\n", dSq, colDistSq);
                     */

                    while(dSq < colDistSq || dSq > colDistSq + contactEpsilonSq) {
                        /* Intelligently choose tCol */
                        tCol = (colDistSq - tLowDSq) / (tHighDSq - tLowDSq);
                        tCol = tLow + tCol * (tHigh - tLow);
                        // tCol = (tCol + tCol + tLow + tHigh) / 4.0;

                        pCol = vector_add(pStart, vector_scale(tCol, v));
                        dSq = dSquared(pCol, pegs[r][c]);

                        if(dSq < colDistSq) {
                          /* fprintf(stderr, "ColSearch High: b=%d t=%f %f(%f) < %f (%f) < %f (%f)\n", i, data->t, tLow, tLowDSq, tCol, dSq, tHigh, tHighDSq); */
                          tHigh = tCol;
                          tHighDSq = dSq;
                        } else if(dSq > colDistSq + contactEpsilonSq) {
                          /* fprintf(stderr, "ColSearch  Low: b=%d t=%f %f(%f) < %f (%f) < %f (%f)\n", i, data->t, tLow, tLowDSq, tCol, dSq, tHigh, tHighDSq); */
                          tLow = tCol;
                          tLowDSq = dSq;
                        }
                    }

                    /* Compute the Normal */
                    nCol = unit_vector(vector_sub(pCol, pegs[r][c]));

                    /* fprintf(stderr, "  tCol=%.6f pCol=(%f, %f) nCol = (%f, %f) ref=(%f, %f) dot(ref,nCol) = %f = dot(v, nCol) = %f\n", tCol, pCol.x, pCol.y, nCol.x, nCol.y, ref.x, ref.y, dot(ref, nCol), dot(v, nCol)); */

                    /* Compute the reflection */
                    reflect(unit_vector(v), nCol, pCol, tCol, simData, data, i);
                }
            }
        }
    }

    /* Check ball/divide collision */
    colDistSq = pegRadius + ballRadius;
    colDistSq *= colDistSq;
    for(i=0; i < simData->nextActive; i++) {
        /* Skip this ball if it isn't down this far yet */
        if(balls[i].center.y > bottom)
            continue;

        for(r = pegRows - 1, c=0; c < pegCols - 1; c++) {
            dSq = pegs[r][c].x - balls[i].center.x;
            dSq *= dSq;
            if(dSq < colDistSq) {
                /* We have a collision */
                    
                /* TODO: Add a constraint that will affect f(x, t) */
                    
                /* Find our path over this time step */
                pStart = simData->lastBalls[i].center;
                v = vector_sub(balls[i].center, pStart);
                    
                /* Intelligent-Binary search to find the point of collision */
                tHigh = 1.0;
                tHighDSq = dSq;
                tLow = 0.0;
                tLowDSq = pegs[r][c].x - pStart.x;
                tLowDSq *= tLowDSq;
                
                /* DEBUGGING: started step interconnected */
                if(tLowDSq < colDistSq) {
                    fprintf(stderr, "Collision (divide) detected at start of timestep.\n");
                }
                
                /*
                 * fprintf(stderr, "collision t=[%.3f,%.3f]: ball %d (%.3f, %.3f), peg %d,%d (%.3f, %.3f)\n", simData->tUpdated, data->t, (int)i, balls[i].center.x, balls[i].center.y, (int)r, (int)c, pegs[r][c].x, pegs[r][c].y);
                 * fprintf(stderr, "    dSq=%f, colDistSq=%f\n", dSq, colDistSq);
                 */
                
                while(dSq < colDistSq || dSq > colDistSq + contactEpsilonSq) {
                    /* Intelligently choose tCol */
                    tCol = (colDistSq - tLowDSq) / (tHighDSq - tLowDSq);
                    tCol = tLow + tCol * (tHigh - tLow);
                    tCol = (tCol + tCol + tLow + tHigh) / 4.0;
                    
                    pCol = vector_add(pStart, vector_scale(tCol, v));
                    dSq = pegs[r][c].x - pCol.x;
                    dSq *= dSq;
                    
                    if(dSq < colDistSq) {
                        /* fprintf(stderr, "ColSearch High: %f(%f) < %f (%f) < %f (%f)\n", tLow, tLowDSq, tCol, dSq, tHigh, tHighDSq); */
                        tHigh = tCol;
                        tHighDSq = dSq;
                    } else if(dSq > colDistSq + contactEpsilonSq) {
                        /* fprintf(stderr, "ColSearch  Low: %f(%f) < %f (%f) < %f (%f)\n", tLow, tLowDSq, tCol, dSq, tHigh, tHighDSq); */
                        tLow = tCol;
                        tLowDSq = dSq;
                    }
                }
                
                /* Compute the Normal */
                nCol.y = 0.0;
                
                /* Check if this is the left side of the peg in which case
                 * we need to flip the normal
                 */
                if(pCol.x < pegs[r][c].x)
                    nCol.x = -1.0;
                else
                    nCol.x = 1.0;
                
                /* fprintf(stderr, "  tCol=%.6f pCol=(%f, %f) nCol = (%f, %f) ref=(%f, %f) dot(ref,nCol) = %f = dot(v, nCol) = %f\n", tCol, pCol.x, pCol.y, nCol.x, nCol.y, ref.x, ref.y, dot(ref, nCol), dot(v, nCol)); */
                
                /* Compute the reflection */
                reflect(unit_vector(v), nCol, pCol, tCol, simData, data, i);
            }
        }
    }
    
    /* Check ball/wall collision */
    colDistSq = ballRadius;
    colDistSq *= colDistSq;
    for(i=0; i < simData->nextActive; i++) {
        for(w = -1; w <= 1; w++) {
            if(w) {
                dSq = w * right - balls[i].center.x;
            } else {
                dSq = ground - balls[i].center.y;
            }
            dSq *= dSq;
            if(dSq < colDistSq) {
                /* We have a collision */
                /* fprintf(stderr, "wall collision: time %f ball %d, wall %d, dSq %f < %f\n", data->t, i, w, dSq, colDistSq); */
                
                /* TODO: Add a constraint that will affect f(x, t) */
                
                /* Find our path over this time step */
                pStart = simData->lastBalls[i].center;
                v = vector_sub(balls[i].center, pStart);
                
                /* Intelligent-Binary search to find the point of collision */
                tHigh = 1.0;
                tHighDSq = dSq;
                tLow = 0.0;

                if(r) {
                    tLowDSq = w * right - pStart.x;
                } else {
                    tLowDSq = ground - pStart.y;
                }
                tLowDSq *= tLowDSq;
                
                /* DEBUGGING: started step interconnected */
                if(tLowDSq < colDistSq - contactEpsilonSq) {
                    fprintf(stderr, "Collision (wall) detected at start of timestep.\n");
                }
                
                /* fprintf(stderr, "collision t=[%.3f,%.3f]: w %d ball %d (%.3f, %.3f), dSq=%f, colDistSq=%f\n", simData->tUpdated, data->t, w, (int)i, balls[i].center.x, balls[i].center.y, dSq, colDistSq); */
                                
                while(dSq < colDistSq || dSq > colDistSq + contactEpsilonSq) {
                    /* Intelligently choose tCol */
                    tCol = (colDistSq - tLowDSq) / (tHighDSq - tLowDSq);
                    tCol = tLow + tCol * (tHigh - tLow);
                    tCol = (tCol + tCol + tLow + tHigh) / 4.0;

                    pCol = vector_add(pStart, vector_scale(tCol, v));
                    if(w) {
                        dSq = w * right - pCol.x;
                    } else {
                        dSq = ground - pCol.y;
                    }
                    dSq *= dSq;
                    
                    if(dSq < colDistSq) {
                        /* fprintf(stderr, "ColSearch High: w=%d b=%d t=%f %f(%f) < %f (%f) < %f (%f)\n", w, i, data->t, tLow, tLowDSq, tCol, dSq, tHigh, tHighDSq); */
                        tHigh = tCol;
                        tHighDSq = dSq;
                    } else if(dSq > colDistSq + contactEpsilonSq) {
                        /* fprintf(stderr, "ColSearch  Low: W=%d b=%d t=%f %f(%f) < %f (%f) < %f (%f)\n", w, i, data->t, tLow, tLowDSq, tCol, dSq, tHigh, tHighDSq); */
                        tLow = tCol;
                        tLowDSq = dSq;
                    }
                }
                
                /* Compute the Normal */
                nCol.x = (ODEFloat)-w;
                nCol.y = 1.0 - (ODEFloat)(w*w);
                
                /* fprintf(stderr, "  tCol=%.6f pCol=(%f, %f) nCol = (%f, %f)\n", tCol, pCol.x, pCol.y, nCol.x, nCol.y); */

                /* Compute the reflection */
                reflect(unit_vector(v), nCol, pCol, tCol, simData, data, i);
            }
        }
    }
        
    /* Spawn new balls */
    while(simData->nextActive < simData->nBalls &&
          simData->initialBalls[simData->nextActive].visible <= data->t) {
        balls[simData->nextActive] = simData->initialBalls[simData->nextActive];
        balls[simData->nextActive].visible = 1.0;

        simData->nextActive++;
    }

    memcpy(simData->lastBalls, balls, simData->nextActive * sizeof(Ball));
    simData->tUpdated = data->t;
}

/* When do we want our balls to spawn.
 * TODO: Have this read from a config file.
 */
static Ball *initBalls(size_t *nBalls) {
    Ball *initialBalls = (Ball *)malloc(numBalls * sizeof(Ball));
    *nBalls = numBalls;

    /* Simple first test case.  One ball with no velocity */
#define initBall(i, v, cx, cy, vx, vy) \
    initialBalls[i].visible = v; \
    initialBalls[i].center.x = cx; \
    initialBalls[i].center.y = cy; \
    initialBalls[i].velocity.x = vx; \
    initialBalls[i].velocity.y = vy;

    initBall(0, 0.0, 0.05, 1.2, 0.1, -0.05);
    initBall(1, 1.0, -0.75, 1.2, 0.04, 0.0);
    initBall(2, 1.0, 0.87, 1.2, -0.02, -0.04);
    initBall(3, 2.0, 0.25, 1.2, 0.07, 0.0);
    initBall(4, 2.0, -0.35, 1.2, 0.1, -0.08);
    initBall(5, 2.0, 0.55, 1.2, -0.5, -1.2);
#undef initBall

    return initialBalls;
}

/* Allocate, Free, Copy SimData */

/* TODO: Make this not dependent on numBalls, so that can be set only on init/copy */
static void *simDataAllocate() {
    SimData *data = (SimData *)malloc(sizeof(SimData));
    if(!data)
        return NULL;

    data->initialBalls = initBalls(&data->nBalls);
    if(!data->initialBalls) {
        free(data);
        return NULL;
    }

    data->lastBalls = (Ball *)malloc(numBalls * sizeof(Ball));
    if(!data->lastBalls) {
        free(data->initialBalls);
        free(data);
        return NULL;
    }

    return (void *)data;
}

static void simDataFree(void *_data) {
    SimData *data = (SimData *)_data;
    if(!data)
        return;

    if(data->initialBalls)
        free(data->initialBalls);

    if(data->lastBalls)
        free(data->lastBalls);

    free(data);
}

static void simDataCopy(void *_dst, const void *_src) {
    SimData *dst = (SimData *)_dst;
    SimData *src = (SimData *)_src;

    dst->tUpdated = src->tUpdated;
    dst->nextActive = src->nextActive;
    memcpy(dst->initialBalls, src->initialBalls, src->nBalls * sizeof(Ball));
    memcpy(dst->lastBalls, src->lastBalls, src->nBalls * sizeof(Ball));
}

static void initPegs() {
    size_t r, c, i;
    /* The rows with an odd number (pegCols - 1) of pegs have gaps between the
     * pegs and walls that are the same size as the gaps between pegs
     */
    ODEFloat hSpacing = (right - left + 0.12) / (ODEFloat)pegCols;

    /* The top is where the 0th row would be, and the bottom is where the last
     * row is
     */
    ODEFloat vSpacing = (top - bottom) / (ODEFloat)pegRows;

    i=0;
    for(r=0; r < pegRows; r++) {
        for(c=0; c < pegCols; c++) {
            /* Even rows don't have the last peg */
            if((r % 2 == 0) && c == (pegCols-1))
                continue;

            pegs[r][c].x = -0.06 + left + (c+1) * hSpacing;
            pegs[r][c].y = top - (r+1) * vSpacing;

            if(r % 2)
                pegs[r][c].x -= hSpacing / 2.0;

            i++;
        }
    }
}

/* Initialize the simulation data. */
static ODEData *initSim() {
    ODEData *data = (ODEData *)calloc(1, sizeof(ODEData));
    SimData *simData;
    if(!data)
        return NULL;

    /* Create our initial conditions for the simulator */
    data->t = 0.0;
    data->simData = simDataAllocate();
    if(!data->simData) {
        free(data);
        return NULL;
    }

    /* For ease of casting... */
    simData = (SimData *)data->simData;

    simData->nextActive = 0;
    simData->tUpdated = 0.0;

    /* C casting rocks... just treat x as a sequence of Balls since Balls are just
     * a sequence of ODEFloats.
     */
    data->x = (ODEFloat *)calloc(simData->nBalls, sizeof(Ball));
    if(!data->x) {
        free(simData->initialBalls);
        free(data->simData);
        free(data);
        return NULL;
    }

    data->dim = simData->nBalls * sizeof(Ball) / sizeof(ODEFloat);

    simDataUpdate(data);

    return data;
}

/* Report dx/dt at the current time */
static void f(const ODEData *data, ODEFloat *f_result) {
    SimData *simData = (SimData *)data->simData;
    Ball *balls = (Ball *)data->x;
    Ball *b_result = (Ball *)f_result;
    size_t i;

    for(i=0; i < simData->nBalls && balls[i].visible; i++) {
        b_result[i].visible = 0.0;
        b_result[i].center.x = balls[i].velocity.x;
        b_result[i].center.y = balls[i].velocity.y;
        b_result[i].velocity.x = 0.0;
        b_result[i].velocity.y = -10.0;
    }
}

int main(int argc, const char **argv) {
    size_t c, r;
    int retval;
    ODEData *data;

    /* Initialize our data */
    initPegs();
    data = initSim();
    if(!data) {
        fprintf(stderr, "Memory allocation error.\n");
        return 2;
    }

    printf("static %u\n", nPegs + pegCols + 2);
    printf("balls %u\n", (unsigned int)((SimData *)data->simData)->nBalls);
    printf("start 0\n");
    printf("end 10000\n");
    printf("step 20\n");

    /* Draw the board */
    for(r=0; r < pegRows; r++) {
        for(c=0; c < pegCols; c++) {
            /* Even rows don't have the last peg */
            if((r % 2 == 0) && c == (pegCols-1))
                continue;

            printf("circle color 0.9 0.9 0.4 1.0\n");
            printf("\tcenter %f %f\n", pegs[r][c].x, pegs[r][c].y);
            printf("\tradius %f\n", pegRadius);
        }
    }

    for(r--, c=0; c < pegCols - 1; c++) {
        printf("polygon color 0.9 0.9 0.4 1.0\n");
        printf("\tverts 4\n");
        printf("\tvert %f %f\n", pegs[r][c].x - pegRadius, pegs[r][c].y);
        printf("\tvert %f %f\n", pegs[r][c].x + pegRadius, pegs[r][c].y);
        printf("\tvert %f %f\n", pegs[r][c].x + pegRadius, ground);
        printf("\tvert %f %f\n", pegs[r][c].x - pegRadius, ground);
    }

    printf("line color 0.1 0.1 0.9\n");
    printf("\tvert %f %f\n", left, ground);
    printf("\tvert %f %f\n", left, top);

    printf("line color 0.1 0.1 0.9\n");
    printf("\tvert %f %f\n", right, ground);
    printf("\tvert %f %f\n", right, top);

    printf("line color 0.1 0.1 0.9\n");
    printf("\tvert %f %f\n", left, ground);
    printf("\tvert %f %f\n", right, ground);

    retval = ode_solver(f, data, ode_step_midpoint, report, simDataUpdate,
                        simDataAllocate, simDataFree, simDataCopy,
                        0.020, 10.0, 0.0005);

    if(retval == 2) {
        fprintf(stderr, "Memory allocation error.\n");
    }
    return retval;
}
